/*
 * Deepkit Framework
 * Copyright (C) 2021 Deepkit UG, Marc J. Schmidt
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the MIT License.
 *
 * You should have received a copy of the MIT License along with this program.
 */
import { DefaultPlatform, isSet, SqlPlaceholderStrategy, typeResolvesToBigInt, typeResolvesToBoolean, typeResolvesToInteger, typeResolvesToNumber, typeResolvesToString, } from '@deepkit/sql';
import { postgresSerializer } from './postgres-serializer.js';
import { isDateType, isReferenceType, isUUIDType, ReflectionClass, ReflectionKind, TypeNumberBrand, } from '@deepkit/type';
import { PostgresSchemaParser } from './postgres-schema-parser.js';
import { PostgreSQLFilterBuilder } from './sql-filter-builder.js';
import { isArray, isObject } from '@deepkit/core';
import sqlstring from 'sqlstring';
function escapeLiteral(value) {
    if (value === null || value === undefined)
        return 'null';
    if (value instanceof Date)
        return sqlstring.escape(value);
    if ('number' === typeof value || 'bigint' === typeof value)
        return String(value);
    if ('string' !== typeof value)
        return escapeLiteral(String(value));
    let hasBackslash = false;
    let escaped = '\'';
    for (let i = 0; i < value.length; i++) {
        const c = value[i];
        if (c === '\'') {
            escaped += c + c;
        }
        else if (c === '\\') {
            escaped += c + c;
            hasBackslash = true;
        }
        else {
            escaped += c;
        }
    }
    escaped += '\'';
    if (hasBackslash) {
        escaped = ' E' + escaped;
    }
    return escaped;
}
export class PostgresPlaceholderStrategy extends SqlPlaceholderStrategy {
    getPlaceholder() {
        return '$' + (++this.offset);
    }
}
export class PostgresPlatform extends DefaultPlatform {
    constructor() {
        super();
        this.defaultSqlType = 'text';
        this.annotationId = 'postgres';
        this.defaultNowExpression = `now()`;
        this.serializer = postgresSerializer;
        this.schemaParserType = PostgresSchemaParser;
        this.placeholderStrategy = PostgresPlaceholderStrategy;
        this.addType(() => true, 'jsonb'); //default everything is jsonb
        this.addType(typeResolvesToNumber, 'double precision');
        this.addType(typeResolvesToInteger, 'integer');
        this.addType(typeResolvesToBigInt, 'bigint');
        this.addType(typeResolvesToBoolean, 'boolean');
        this.addType(typeResolvesToString, 'text');
        this.addType(type => type.kind === ReflectionKind.number && type.brand === TypeNumberBrand.integer, 'integer');
        this.addType(type => type.kind === ReflectionKind.number && type.brand === TypeNumberBrand.int8, 'smallint');
        this.addType(type => type.kind === ReflectionKind.number && type.brand === TypeNumberBrand.uint8, 'smallint');
        this.addType(type => type.kind === ReflectionKind.number && type.brand === TypeNumberBrand.int16, 'smallint');
        this.addType(type => type.kind === ReflectionKind.number && type.brand === TypeNumberBrand.uint16, 'smallint');
        this.addType(type => type.kind === ReflectionKind.number && type.brand === TypeNumberBrand.int32, 'integer');
        this.addType(type => type.kind === ReflectionKind.number && type.brand === TypeNumberBrand.uint32, 'integer');
        this.addType(type => type.kind === ReflectionKind.number && type.brand === TypeNumberBrand.float32, 'real');
        this.addType(type => type.kind === ReflectionKind.number && type.brand === TypeNumberBrand.float64, 'double precision');
        this.addType(type => type.kind === ReflectionKind.number && type.brand === TypeNumberBrand.float, 'double precision');
        this.addType(isUUIDType, 'uuid');
        this.addBinaryType('bytea');
        this.addType(isDateType, 'timestamp');
    }
    getSqlTypeCaster(type) {
        if (isReferenceType(type)) {
            type = ReflectionClass.from(type).getPrimary().type;
        }
        const dbType = this.getTypeMapping(type);
        if (!dbType)
            return super.getSqlTypeCaster(type);
        return (placeholder) => placeholder + '::' + dbType.sqlType;
    }
    getAggregateSelect(tableName, property, func) {
        if (func === 'group_concat') {
            return `array_to_string(array_agg(${tableName}.${this.quoteIdentifier(property.name)}), ',')`;
        }
        return super.getAggregateSelect(tableName, property, func);
    }
    createSqlFilterBuilder(adapter, schema, tableName) {
        return new PostgreSQLFilterBuilder(adapter, schema, tableName, this.serializer, new this.placeholderStrategy);
    }
    getDeepColumnAccessor(table, column, path) {
        return `${table ? table + '.' : ''}${this.quoteIdentifier(column)}->${this.quoteValue(path)}`;
    }
    quoteValue(value) {
        if (!(value instanceof Date) && (isObject(value) || isArray(value)))
            return escapeLiteral(JSON.stringify(value));
        if (value instanceof Date)
            return 'TIMESTAMP ' + sqlstring.escape(value);
        return escapeLiteral(value);
    }
    getColumnType(column) {
        const ddl = [];
        if (column.isAutoIncrement) {
            ddl.push(`SERIAL`);
        }
        else {
            const foreignKey = column.table.getForeignKeyOfLocalColumn(column);
            if (foreignKey) {
                const [foreignPk] = foreignKey.foreignColumns;
                if (foreignPk.isAutoIncrement) {
                    //the foreignPK has `SERIAL` type, which is basically an INTEGER.
                    ddl.push('INTEGER');
                }
                else {
                    ddl.push((foreignPk.type || 'INTEGER') + column.getSizeDefinition());
                }
            }
            else {
                ddl.push((column.type || 'INTEGER') + column.getSizeDefinition());
            }
        }
        return ddl.filter(isSet).join(' ');
    }
    getColumnDDL(column) {
        const ddl = [];
        ddl.push(this.getIdentifier(column));
        ddl.push(this.getColumnType(column));
        if (!column.isAutoIncrement) {
            ddl.push(column.isNotNull ? this.getNotNullString() : this.getNullString());
            ddl.push(this.getColumnDefaultValueDDL(column));
        }
        return ddl.filter(isSet).join(' ');
    }
    getModifyColumnDDL(diff) {
        // postgres doesn't support multiple column modifications in one ALTER TABLE statement
        // see https://www.postgresql.org/docs/current/sql-altertable.html
        const lines = [];
        const identifier = this.getIdentifier(diff.to);
        if (diff.from.type !== diff.to.type || diff.from.isAutoIncrement !== diff.to.isAutoIncrement) {
            lines.push(`ALTER TABLE ${this.getIdentifier(diff.to.table)} ALTER ${identifier} TYPE ${this.getColumnType(diff.to)}`);
        }
        if (diff.from.isNotNull !== diff.to.isNotNull) {
            if (diff.to.defaultExpression !== undefined || diff.to.defaultValue !== undefined) {
                lines.push(`ALTER TABLE ${this.getIdentifier(diff.to.table)} ALTER ${identifier} SET ${this.getColumnDefaultValueDDL(diff.to)}`);
            }
            else {
                lines.push(`ALTER TABLE ${this.getIdentifier(diff.to.table)} ALTER ${identifier} DROP DEFAULT`);
            }
        }
        if (diff.from.isNotNull !== diff.to.isNotNull) {
            if (diff.to.isNotNull) {
                //NOT NULL is newly added, so we need to update all existing rows to have a value
                const defaultExpression = this.getDefaultExpression(diff.to);
                if (defaultExpression) {
                    lines.push(`UPDATE ${this.getIdentifier(diff.to.table)} SET ${identifier} = ${defaultExpression} WHERE ${identifier} IS NULL`);
                }
                lines.push(`ALTER TABLE ${this.getIdentifier(diff.to.table)} ALTER ${identifier} SET NOT NULL`);
            }
            else {
                lines.push(`ALTER TABLE ${this.getIdentifier(diff.to.table)} ALTER ${identifier} DROP NOT NULL`);
            }
        }
        return lines.join(';\n');
    }
    getDefaultExpression(column) {
        if (undefined !== column.defaultValue) {
            // if type is from JSON type, we need to cast whatever is in defaultValue to JSON
            if (column.type === 'jsonb')
                return `${this.quoteValue(JSON.stringify(column.defaultValue))}::jsonb`;
            if (column.type === 'json')
                return `${this.quoteValue(JSON.stringify(column.defaultValue))}::json`;
        }
        return super.getDefaultExpression(column);
    }
    getUniqueDDL(unique) {
        return `CONSTRAINT ${this.getIdentifier(unique)} UNIQUE (${this.getColumnListDDL(unique.columns)})`;
    }
    getDropIndexDDL(index) {
        if (index.isUnique) {
            return `ALTER TABLE ${this.getIdentifier(index.table)} DROP CONSTRAINT ${this.getIdentifier(index)}`;
        }
        return super.getDropIndexDDL(index);
    }
    supportsInlineForeignKey() {
        return false;
    }
    supportsAggregatedAlterTable() {
        return false;
    }
    supportsSelectFor() {
        return true;
    }
    getAutoIncrement() {
        return '';
    }
    getDropTableDDL(table) {
        return `DROP TABLE IF EXISTS ${this.getIdentifier(table)} CASCADE`;
    }
    getUseSchemaDDL(table) {
        if (!table.schemaName)
            return '';
        return `SET search_path TO ${this.quoteIdentifier(table.schemaName)}`;
    }
    getResetSchemaDDL(table) {
        if (!table.schemaName)
            return '';
        return `SET search_path TO public`;
    }
}
//# sourceMappingURL=postgres-platform.js.map